"""Module to map risk-estimates to a variety of evaluation metrics.

Notes
-----
Code based on the `error_parity.evaluation` module,
at: https://github.com/socialfoundations/error-parity/blob/main/error_parity/evaluation.py
"""
from __future__ import annotations

import logging
import statistics
from typing import Callable, Optional

import numpy as np
from netcal.metrics import ECE
from sklearn.metrics import brier_score_loss, confusion_matrix, log_loss, roc_auc_score, roc_curve

from ._utils import is_valid_number, join_dictionaries, safe_division


def evaluate_binary_predictions(y_true: np.ndarray, y_pred: np.ndarray) -> dict:
    """Evaluates the provided binary predictions on common performance metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The binary predictions.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # Compute confusion matrix
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=(0, 1)).ravel()
    tn, fp, fn, tp = map(int, (tn, fp, fn, tp))

    total = tn + fp + fn + tp
    pred_pos = tp + fp
    pred_neg = tn + fn
    assert pred_pos + pred_neg == total

    label_pos = tp + fn
    label_neg = tn + fp
    assert label_pos + label_neg == total

    results = {}

    # Accuracy
    results["accuracy"] = (tp + tn) / total

    # True Positive Rate (Recall)
    results["tpr"] = safe_division(tp, label_pos, worst_result=0)

    # False Negative Rate (1 - TPR)
    results["fnr"] = safe_division(fn, label_pos, worst_result=1)
    assert results["tpr"] + results["fnr"] == 1

    # False Positive Rate
    results["fpr"] = safe_division(fp, label_neg, worst_result=1)

    # True Negative Rate
    results["tnr"] = safe_division(tn, label_neg, worst_result=0)
    assert results["tnr"] + results["fpr"] == 1

    # Balanced accuracy
    results["balanced_accuracy"] = 0.5 * (results["tpr"] + results["tnr"])

    # Precision
    results["precision"] = safe_division(tp, pred_pos, worst_result=0)

    # Positive Prediction Rate
    results["ppr"] = safe_division(pred_pos, total, worst_result=0)

    return results


def evaluate_binary_predictions_fairness(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sensitive_attribute: np.ndarray,
    return_groupwise_metrics: bool = False,
    min_group_size: float = 0.04,
) -> dict:
    """Evaluates fairness of the given predictions.

    Fairness metrics are computed as the ratios between group-wise performance
    metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The discretized predictions.
    sensitive_attribute : np.ndarray
        The sensitive attribute (protected group membership).
    return_groupwise_metrics : bool, optional
        Whether to return group-wise performance metrics (bool: True) or only
        the ratios between these metrics (bool: False), by default False.
    min_group_size : float, optional
        The minimum fraction of samples (as a fraction of the total number of
        samples) that a group must have to be considered for fairness
        evaluation, by default 0.04. This is meant to avoid evaluating metrics
        on very small groups which leads to noisy and inconsistent results.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # All unique values for the sensitive attribute
    unique_groups = np.unique(sensitive_attribute)

    results = {}
    groupwise_metrics = {}
    unique_metrics = set()

    # Helper to compute key/name of a group-wise metric
    def group_metric_name(metric_name, group_name):
        return f"{metric_name}_group={group_name}"

    if len(unique_groups) <= 1:
        logging.error(f"Found a single unique sensitive attribute: {unique_groups}")
        return {}

    for s_value in unique_groups:
        # Indices of samples that belong to the current group
        group_indices = np.argwhere(sensitive_attribute == s_value).flatten()

        if len(group_indices) < min_group_size * len(y_true):
            logging.warning(f"Skipping group {s_value} with {len(group_indices)} samples")
            continue

        # Filter labels and predictions for samples of the current group
        group_labels = y_true[group_indices]
        group_preds = y_pred[group_indices]

        # Evaluate group-wise performance
        curr_group_metrics = evaluate_binary_predictions(group_labels, group_preds)

        # Add group-wise metrics to the dictionary
        groupwise_metrics.update(
            {
                group_metric_name(metric_name, s_value): metric_value
                for metric_name, metric_value in curr_group_metrics.items()
            }
        )

        unique_metrics = unique_metrics.union(curr_group_metrics.keys())

    # Compute ratios and absolute diffs
    for metric_name in unique_metrics:
        curr_metric_results = [
            groupwise_metrics[curr_group_metric_name]
            for group_name in unique_groups
            if (curr_group_metric_name := group_metric_name(metric_name, group_name)) in groupwise_metrics
        ]

        # Metrics' ratio
        ratio_name = f"{metric_name}_ratio"

        # NOTE: should this ratio be computed w.r.t. global performance?
        # - i.e., min(curr_metric_results) / global_curr_metric_result;
        # - same question for the absolute diff calculations;
        results[ratio_name] = safe_division(
            min(curr_metric_results), max(curr_metric_results),
            worst_result=0,
        )

        # Metrics' absolute difference
        diff_name = f"{metric_name}_diff"
        results[diff_name] = max(curr_metric_results) - min(curr_metric_results)

    # ** Equalized odds **
    # default value: use maximum constraint violation for TPR and FPR equality
    results["equalized_odds_ratio"] = min(
        results["fnr_ratio"],
        results["fpr_ratio"],
    )
    results["equalized_odds_diff"] = max(
        results["tpr_diff"],  # same as FNR diff
        results["fpr_diff"],  # same as TNR diff
    )

    # Optionally, return group-wise metrics as well
    if return_groupwise_metrics:
        results.update(groupwise_metrics)

    return results


def compute_best_threshold(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    false_pos_cost: float = 1.0,
    false_neg_cost: float = 1.0,
) -> float:
    """Computes the binarization threshold that maximizes accuracy.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred_scores : np.ndarray
        The predicted risk scores.
    false_pos_cost : float, optional
        The cost of a false positive error, by default 1.0
    false_neg_cost : float, optional
        The cost of a false negative error, by default 1.0

    Returns
    -------
    best_threshold : float
        The threshold value that maximizes accuracy for the given predictions.
    """

    # Compute TPR and FPR for all possible thresholds
    fpr, tpr, thresholds = roc_curve(y_true, y_pred_scores)

    # Compute the cost of each threshold
    costs = false_pos_cost * fpr + false_neg_cost * (1 - tpr)

    # Get the threshold that minimizes the cost
    best_threshold = thresholds[np.argmin(costs)]
    return float(best_threshold)


def evaluate_predictions(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    sensitive_attribute: np.ndarray = None,
    threshold: float | str = "best",
    model_name: str = None,
) -> dict:
    """Evaluates predictions on common performance and fairness metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred_scores : np.ndarray
        The predicted scores.
    sensitive_attribute: np.ndarray, optional
        The sensitive attribute data. Will compute fairness metrics if provided.
    threshold : float | str, optional
        The threshold to use for binarizing the predictions, or "best" to infer
        which threshold maximizes accuracy.
    model_name : str, optional
        The name of the model to be used on the plots, by default None.

    Returns
    -------
    results : dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # Compute threshold if necessary
    if threshold == "best":
        threshold = compute_best_threshold(y_true, y_pred_scores)
    assert is_valid_number(threshold) and 0 <= threshold <= 1, \
        f"Invalid threshold: {threshold}"

    # Save initial results' statistics
    results = {
        "threshold": threshold,
        "n_samples": len(y_true),
        "n_positives": np.sum(y_true).item(),
        "n_negatives": len(y_true) - np.sum(y_true).item(),
        "model_name": model_name,
    }

    # Compute binary predictions using the default threshold
    y_pred_binary = (y_pred_scores >= threshold).astype(int)

    # Evaluate binary predictions
    results.update(evaluate_binary_predictions(y_true, y_pred_binary))

    # Add loss functions as proxies for calibration
    results["log_loss"] = log_loss(y_true, y_pred_scores, labels=[0, 1])
    results["brier_score_loss"] = float(brier_score_loss(y_true, y_pred_scores))

    # Evaluate fairness metrics
    if sensitive_attribute is not None:
        results.update(evaluate_binary_predictions_fairness(y_true, y_pred_binary, sensitive_attribute))

    # Compute additional metrics
    results["roc_auc"] = float(roc_auc_score(y_true, y_pred_scores))

    # Compute Expected Calibration Error
    # TODO: re-implement ECE scorer to avoid including 10 other dependencies for this one metric...
    class_preds = np.stack([1 - y_pred_scores, y_pred_scores], axis=1)
    n_bins = 10
    results["ece"] = ECE(bins=n_bins, equal_intervals=True).measure(class_preds, y_true)

    # Try to compute ECE with quantile bins
    try:
        results["ece_quantile"] = ECE(bins=n_bins, equal_intervals=False).measure(class_preds, y_true)
    except Exception as err:
        logging.warning(f"Failed to compute ECE quantile: {err}")
        results["ece_quantile"] = None

    return results


def bootstrap_estimate(
    eval_func: Callable[[np.ndarray, np.ndarray, np.ndarray], dict[str, float]],
    *,
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    sensitive_attribute: np.ndarray = None,
    k: int = 200,
    confidence_pct: float = 95,
    seed: int = 42,
) -> dict:
    """Computes bootstrap estimates of the given evaluation function.

    Parameters
    ----------
    eval_func : Callable[[np.ndarray, np.ndarray, np.ndarray], dict[str, float]]
        The evaluation function to run for each bootstrap sample. Must follow
        the signature `eval_func(y_true, y_pred_scores, sensitive_attribute)`.
    y_true : np.ndarray
        The true labels.
    y_pred_scores : np.ndarray
        The predicted scores.
    sensitive_attribute : np.ndarray, optional
        Optionally, provide the sensitive attribute data to compute fairness
        metrics, by default None.
    k : int, optional
        How many bootstrap samples to draw, by default 200.
    confidence_pct : float, optional
        The confidence interval to use, in percentage, by default 95.
    seed : int, optional
        The random seed, by default 42.

    Returns
    -------
    results : dict[str, float]
        A dictionary containing bootstrap estimates for a variety of metrics.
    """
    assert len(y_true) == len(y_pred_scores)
    rng = np.random.default_rng(seed=seed)

    # Draw k bootstrap samples with replacement
    results = []
    for _ in range(k):
        # Indices of current bootstrap sample
        indices = rng.choice(len(y_true), replace=True, size=len(y_true))

        # Evaluate predictions on this bootstrap sample
        results.append(
            eval_func(
                y_true[indices],            # ground-truth labels
                y_pred_scores[indices],     # predicted risk scores
                sensitive_attribute[indices]
                if sensitive_attribute is not None else None,  # sensitive attributes
            )
        )

    # Compute statistics from bootstrapped results
    all_metrics = set(results[0].keys())

    bt_mean = {}
    bt_stdev = {}
    bt_percentiles = {}

    low_percentile = (100 - confidence_pct) / 2
    confidence_percentiles = [low_percentile, 100 - low_percentile]

    for m in all_metrics:
        metric_values = [r[m] for r in results]

        bt_mean[m] = statistics.mean(metric_values)
        bt_stdev[m] = statistics.stdev(metric_values)
        bt_percentiles[m] = tuple(np.percentile(metric_values, confidence_percentiles))

    # Join bootstrap results for all metrics
    return join_dictionaries(
        *(
            {
                f"{metric}_mean": bt_mean[metric],
                f"{metric}_stdev": bt_stdev[metric],
                f"{metric}_low-percentile": bt_percentiles[metric][0],
                f"{metric}_high-percentile": bt_percentiles[metric][1],
            }
            for metric in sorted(bt_mean.keys())
        )
    )


def evaluate_predictions_bootstrap(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    sensitive_attribute: np.ndarray = None,
    threshold: float | str = "best",
    k: int = 200,
    confidence_pct: float = 95,
    seed: int = 42,
) -> dict:
    """Computes bootstrap estimates of classification metrics for the given predictions.

    Parameters
    ----------
    y_true : np.ndarray
        The true labels.
    y_pred_scores : np.ndarray
        The score predictions.
    sensitive_attribute : np.ndarray, optional
        The sensitive attribute data. Will compute fairness metrics if provided.
    threshold : float | str, optional
        The threshold to use for binarizing the predictions, or "best" to infer
        which threshold maximizes accuracy, by default "best".
    k : int, optional
        How many bootstrap samples to draw, by default 200.
    confidence_pct : float, optional
        How large of a confidence interval to use when reporting lower and upper
        bounds, by default 95 (i.e., 2.5 to 97.5 percentile of results).
    seed : int, optional
        The random seed, by default 42.

    Returns
    -------
    results : dict[str, float]
        A dictionary containing bootstrap estimates for a variety of metrics.
    """
    return bootstrap_estimate(
        eval_func=lambda labels, scores, sens_attr=None: evaluate_predictions(
            y_true=labels,
            y_pred_scores=scores,
            sensitive_attribute=sens_attr,
            threshold=threshold,
        ),
        y_true=y_true,
        y_pred_scores=y_pred_scores,
        sensitive_attribute=sensitive_attribute,
        k=k,
        confidence_pct=confidence_pct,
        seed=seed,
    )



"""Definition of a generic TaskMetadata class.
"""
from __future__ import annotations

import dataclasses
import logging
from dataclasses import dataclass, field
from typing import Callable, ClassVar, Iterable

import pandas as pd

from ._utils import hash_dict
from .col_to_text import ColumnToText
from .qa_interface import DirectNumericQA, MultipleChoiceQA, QAInterface
from .threshold import Threshold


@dataclass
class TaskMetadata:
    """A base class to hold information on a prediction task."""

    name: str
    """The name of the task."""

    features: list[str]
    """The names of the features used in the task."""

    target: str
    """The name of the target column."""

    cols_to_text: dict[str, ColumnToText]
    """A mapping between column names and their textual descriptions."""

    sensitive_attribute: str = None
    """The name of the column used as the sensitive attribute data (if provided)."""

    target_threshold: Threshold = None
    """The threshold used to binarize the target column (if provided)."""

    multiple_choice_qa: MultipleChoiceQA = None
    """The multiple-choice question and answer interface for this task."""

    direct_numeric_qa: DirectNumericQA = None
    """The direct numeric question and answer interface for this task."""

    description: str = None
    """A description of the task, including the population to which the task pertains to."""

    _use_numeric_qa: bool = False
    """Whether to use numeric Q&A instead of multiple-choice Q&A prompts. Default is False."""

    # Class-level task storage
    _tasks: ClassVar[dict[str, "TaskMetadata"]] = field(default={}, init=False, repr=False)

    def __post_init__(self):
        # Check if this task had already been created
        if self.name in TaskMetadata._tasks:
            logging.error(f"A task with `name='{self.name}'` already exists. Overwriting...")

        # Add this task to the class-level dictionary
        TaskMetadata._tasks[self.name] = self

        # Check all required columns are provided by the `cols_to_text` map
        self.check_task_columns_are_available(self.cols_to_text.keys())

        # Check target is provided
        if self.target is None:
            logging.warning(
                f"No target column provided for task '{self.name}'. "
                f"Will not be able to generate predictions or use task Q&A prompts. "
                f"Will still be able to generate row descriptions."
            )
            return

        # If no question is explicitly provided, use the question from the target column
        if self.multiple_choice_qa is None and self.direct_numeric_qa is None and self.target is not None:
            logging.warning(
                f"No question was explicitly provided for task '{self.name}'. "
                f"Inferring from target column's default question ({self.get_target()}).")

            if self.cols_to_text[self.get_target()]._question is not None:
                question = self.cols_to_text[self.get_target()]._question
                self.set_question(question)

        # Make sure Q&A related attributes are consistent
        if (
            self._use_numeric_qa is True and self.direct_numeric_qa is None
            or self._use_numeric_qa is False and self.multiple_choice_qa is None
        ):
            raise ValueError("Inconsistent Q&A attributes provided.")

    def __hash__(self) -> int:
        hashable_params = dataclasses.asdict(self)
        hashable_params.pop("cols_to_text")
        hashable_params["question_hash"] = hash(self.question)
        return int(hash_dict(hashable_params), 16)

    def check_task_columns_are_available(
        self,
        available_cols: list[str],
        raise_: bool = True,
    ) -> bool:
        """Checks if all columns required by this task are available.

        Parameters
        ----------
        available_cols : list[str]
            The list of column names available in the dataset.
        raise_ : bool, optional
            Whether to raise an error if some columns are missing, by default True.

        Returns
        -------
        all_available : bool
            True if all required columns are present in the given list of
            available columns, False otherwise.
        """
        required_cols = self.features + ([self.get_target()] if self.target else [])
        missing_cols = set(required_cols) - set(available_cols)

        if raise_ and len(missing_cols) > 0:
            raise ValueError(
                f"The following required task columns were not found in the dataset: "
                f"{list(missing_cols)};"
            )

        return len(missing_cols) == 0   # Return True if all columns are present

    def get_target(self) -> str:
        """Resolves the name of the target column depending on `self.target_threshold`."""
        if self.target is None:
            logging.critical(f"No target column provided for task {self.name}.")
            return None

        if self.target_threshold is None:
            return self.target
        else:
            return self.target_threshold.apply_to_column_name(self.target)

    def set_question(self, question: QAInterface):
        """Sets the Q&A interface for this task."""
        logging.info(f"Setting question for task '{self.name}' to '{question.text}'.")

        if isinstance(question, MultipleChoiceQA):
            self.multiple_choice_qa = question
            self._use_numeric_qa = False
        elif isinstance(question, DirectNumericQA):
            self.direct_numeric_qa = question
            self._use_numeric_qa = True
        else:
            raise ValueError(f"Invalid question type: {type(question).__name__}")

    @property
    def use_numeric_qa(self) -> bool:
        """Getter for whether to use numeric Q&A instead of multiple-choice Q&A prompts."""
        return self._use_numeric_qa

    @use_numeric_qa.setter
    def use_numeric_qa(self, use_numeric_qa: bool):
        """Setter for whether to use numeric Q&A instead of multiple-choice Q&A prompts."""
        logging.info(
            f"Changing Q&A mode for task '{self.name}' to "
            f"{'numeric' if use_numeric_qa else 'multiple-choice'}.")
        self._use_numeric_qa = use_numeric_qa

    @classmethod
    def get_task(cls, name: str, use_numeric_qa: bool = False) -> TaskMetadata:
        """Fetches a previously created task by its name.

        Parameters
        ----------
        name : str
            The name of the task to fetch.
        use_numeric_qa : bool, optional
            Whether to set the retrieved task to use verbalized numeric Q&A
            instead of the default multiple-choice Q&A prompts. Default is False.

        Returns
        -------
        task : TaskMetadata
            The task object with the given name.

        Raises
        ------
        ValueError
            Raised if the task with the given name has not been created yet.
        """
        if name not in cls._tasks:
            raise ValueError(f"Task '{name}' has not been created yet.")

        # Retrieve the task object
        task = cls._tasks[name]

        # Set Q&A interface type
        task.use_numeric_qa = use_numeric_qa

        return task

    @property
    def question(self) -> QAInterface:
        """Getter for the Q&A interface for this task."""

        # Resolve direct numeric Q&A vs multiple-choice Q&A
        if self._use_numeric_qa:
            q = self.direct_numeric_qa
        else:
            q = self.multiple_choice_qa

        if q is None:
            logging.critical(f"No Q&A interface provided for task {self.name}.")
        return q

    def get_row_description(self, row: pd.Series) -> str:
        """Encode a description of a given data row in textual form."""
        row = row[self.features]
        return (
            "\n".join(
                "- " + self.cols_to_text[col].get_text(val)
                for col, val in row.items()
            )
        )

    def sensitive_attribute_value_map(self) -> Callable:
        """Returns a mapping between sensitive attribute values and their descriptions."""
        if self.sensitive_attribute is None:
            logging.warning("No sensitive attribute provided for this task.")
            return {}
        return self.cols_to_text[self.sensitive_attribute].value_map

    def create_task_with_feature_subset(self, feature_subset: Iterable[str]):
        """Creates a new task with a subset of the original features."""
        # Convert iterable to list
        feature_subset = list(feature_subset)

        # Check if features are a subset of the original features
        if not set(feature_subset).issubset(self.features):
            raise ValueError(
                f"`feature_subset` must be a subset of the original features; "
                f"following features are not in the original set: "
                f"{set(feature_subset) - set(self.features)}"
            )

        # Return new TaskMetadata object
        return dataclasses.replace(
            self,
            name=f"{self.name}_" + "_".join(sorted(feature_subset)),
            features=feature_subset,
        )



"""A benchmark class for measuring and evaluating LLM calibration.
"""
from __future__ import annotations

import dataclasses
import logging
from functools import partial
from pathlib import Path
from typing import Optional

import numpy as np
from transformers import AutoModelForCausalLM, AutoTokenizer

from ._io import load_json, save_json
from ._utils import hash_dict, is_valid_number, get_current_timestamp
from .acs.acs_dataset import ACSDataset
from .acs.acs_tasks import ACSTaskMetadata
from .classifier import LLMClassifier, TransformersLLMClassifier, WebAPILLMClassifier
from .dataset import Dataset
from .evaluation import evaluate_predictions
from .plotting import render_evaluation_plots, render_fairness_plots
from .prompting import encode_row_prompt, encode_row_prompt_few_shot
from .task import TaskMetadata

DEFAULT_SEED = 42
DEFAULT_FIT_THRESHOLD_N = 100
DEFAULT_ROOT_RESULTS_DIR = Path(".")


@dataclasses.dataclass(frozen=True, eq=True)
class BenchmarkConfig:
    """A dataclass to hold the configuration for risk-score benchmark.

    Attributes
    ----------
    numeric_risk_prompting : bool, optional
        Whether to prompt for numeric risk-estimates instead of multiple-choice
        Q&A, by default False.
    few_shot : int | None, optional
        Whether to use few-shot prompting with a given number of examples, by
        default None.
    reuse_few_shot_examples : bool, optional
        Whether to reuse the same samples for few-shot prompting (or sample new
        ones every time), by default False.
    batch_size : int | None, optional
        The batch size to use for inference.
    context_size : int | None, optional
        The maximum context size when prompting the LLM.
    correct_order_bias : bool, optional
        Whether to correct the ordering bias in multiple-choice Q&A when
        prompting the LLM, by default True.
    feature_subset : list[str] | None, optional
        Whether to use a subset of the standard feature set for the task. The
        list should contain the names of the columns of features to use.
    population_filter : dict | None, optional
        Optional population filter for this benchmark; must follow the format
        `{"column_name": "value"}`.
    seed : int, optional
        Random seed -- to set for reproducibility.
    """

    numeric_risk_prompting: bool = False
    few_shot: int | None = None
    reuse_few_shot_examples: bool = False
    batch_size: int | None = None
    context_size: int | None = None
    correct_order_bias: bool = True
    feature_subset: list[str] | None = None
    population_filter: dict | None = None
    seed: int = DEFAULT_SEED

    @classmethod
    def default_config(cls, **changes):
        """Returns the default configuration with optional changes."""
        return cls(**changes)

    def update(self, **changes) -> BenchmarkConfig:
        """Update the configuration with new values."""
        possible_keys = dataclasses.asdict(self).keys()
        valid_changes = {k: v for k, v in changes.items() if k in possible_keys}

        # Log config changes
        if valid_changes:
            logging.info(f"Updating benchmark configuration with: {valid_changes}")

        # Log unused kwargs
        if len(valid_changes) < len(changes):
            unused_kwargs = {k: v for k, v in changes.items() if k not in possible_keys}
            logging.warning(f"Unused config arguments: {unused_kwargs}")

        return dataclasses.replace(self, **valid_changes)

    @classmethod
    def load_from_disk(cls, path: str | Path):
        """Load the configuration from disk."""
        obj = load_json(path)
        if isinstance(obj, dict):
            return cls(**obj)
        else:
            raise ValueError(f"Invalid configuration file '{path}'.")

    def save_to_disk(self, path: str | Path):
        """Save the configuration to disk."""
        save_json(dataclasses.asdict(self), path)

    def __hash__(self) -> int:
        """Generates a unique hash for the configuration."""
        cfg = dataclasses.asdict(self)
        cfg["feature_subset"] = tuple(cfg["feature_subset"]) if cfg["feature_subset"] else None
        cfg["population_filter_hash"] = (
            hash_dict(cfg["population_filter"])
            if cfg["population_filter"] else None
        )
        return int(hash_dict(cfg), 16)


class Benchmark:
    """Measures and evaluates risk scores produced by an LLM."""

    """
    Standardized configurations for the ACS data to use for benchmarking.
    """
    ACS_DATASET_CONFIGS = {
        # ACS survey configs
        "survey_year": "2018",
        "horizon": "1-Year",
        "survey": "person",

        # Data split configs
        "test_size": 0.1,
        "val_size": 0.1,
        "subsampling": None,

        # Fixed random seed
        "seed": 42,
    }

    def __init__(
        self,
        llm_clf: LLMClassifier,
        dataset: Dataset,
        config: BenchmarkConfig = BenchmarkConfig.default_config(),
    ):
        """A benchmark object to measure and evaluate risk scores produced by an LLM.

        Parameters
        ----------
        llm_clf : LLMClassifier
            A language model classifier object (can be local or web-hosted).
        dataset : Dataset
            The dataset object to use for the benchmark.÷
        config : BenchmarkConfig, optional
            The configuration object used to create the benchmark parameters.
            **NOTE**: This is used to uniquely identify the benchmark object for
            reproducibility; it **will not be used to change the benchmark
            behavior**. To configure the benchmark, pass a configuration object
            to the Benchmark.make_benchmark method.
        """
        self.llm_clf = llm_clf
        self.dataset = dataset
        self.config = config

        self._y_test_scores: Optional[np.ndarray] = None
        self._results_root_dir: Optional[Path] = DEFAULT_ROOT_RESULTS_DIR
        self._results: Optional[dict] = None
        self._plots: Optional[dict] = None

        # Log initialization
        msg = (
            f"\n** Benchmark initialization **\n"
            f"Model: {self.model_name};\n"
            f"Task: {self.task.name};\n"
            f"Hash: {hash(self)};\n"
        )
        logging.info(msg)

    @property
    def configs_dict(self) -> dict:
        cnf = dataclasses.asdict(self.config)

        # Add info on model, task, and dataset
        cnf["model_name"] = self.model_name
        cnf["model_hash"] = hash(self.llm_clf)
        cnf["task_name"] = self.task.name
        cnf["task_hash"] = hash(self.task)
        cnf["dataset_name"] = self.dataset.name
        cnf["dataset_subsampling"] = self.dataset.subsampling
        cnf["dataset_hash"] = hash(self.dataset)

        return cnf

    @property
    def results(self):
        # Add benchmark configs to the results
        self._results["config"] = self.configs_dict
        self._results["benchmark_hash"] = hash(self)
        self._results["results_dir"] = self.results_dir.as_posix()
        self._results["results_root_dir"] = self.results_root_dir.as_posix()
        self._results["current_time"] = get_current_timestamp()

        return self._results

    @property
    def task(self):
        return self.llm_clf.task

    @property
    def model_name(self):
        return self.llm_clf.model_name

    @property
    def results_root_dir(self) -> Path:
        return Path(self._results_root_dir)

    @results_root_dir.setter
    def results_root_dir(self, path: str | Path):
        self._results_root_dir = Path(path).expanduser().resolve()

    @property
    def results_dir(self) -> Path:
        """Get the results directory for this benchmark."""
        # Get subfolder name
        subfolder_name = f"{self.model_name}_bench-{hash(self)}"
        subfolder_dir_path = Path(self.results_root_dir) / subfolder_name

        # Create subfolder directory if it does not exist
        subfolder_dir_path.mkdir(exist_ok=True, parents=True)
        return subfolder_dir_path

    def __hash__(self) -> int:
        hash_params = dict(
            llm_clf_hash=hash(self.llm_clf),
            dataset_hash=hash(self.dataset),
            config_hash=hash(self.config),
        )

        return int(hash_dict(hash_params), 16)

    def _get_predictions_save_path(self, data_split: str) -> Path:
        """Standardized path to file containing predictions for the given data split."""
        assert data_split in ("train", "val", "test")
        return self.results_dir / f"{self.dataset.name}.{data_split}_predictions.csv"

    def run(self, results_root_dir: str | Path, fit_threshold: int | bool = 0) -> float:
        """Run the calibration benchmark experiment.

        Parameters
        ----------
        results_root_dir : str | Path
            Path to root directory under which results will be saved.
        fit_threshold : int | bool, optional
            Whether to fit the binarization threshold on a given number of
            training samples, by default 0 (will not fit the threshold).

        Returns
        -------
        float
            The benchmark metric value. By default this is the ECE score.
        """
        if self._results is not None:
            logging.warning("Benchmark was already run. Overriding previous results.")

        # Update results directory
        self.results_root_dir = Path(results_root_dir)

        # Get test data
        X_test, y_test = self.dataset.get_test()
        logging.info(f"Test data features shape: {X_test.shape}")

        # Get sensitive attribute data if available
        s_test = None
        if self.task.sensitive_attribute is not None:
            s_test = self.dataset.get_sensitive_attribute_data().loc[y_test.index]

        # Get LLM risk-estimate predictions for each row in the test set
        test_predictions_save_path = self._get_predictions_save_path("test")
        self._y_test_scores = self.llm_clf.predict_proba(
            data=X_test,
            predictions_save_path=test_predictions_save_path,
            labels=y_test,  # used only to save alongside predictions in disk
        )
        self._y_test_scores = self.llm_clf._get_positive_class_scores(self._y_test_scores)

        # If requested, fit the threshold on a small portion of the train set
        if fit_threshold:
            if fit_threshold is True:
                fit_threshold = DEFAULT_FIT_THRESHOLD_N
            elif not is_valid_number(fit_threshold) or fit_threshold <= 0:
                raise ValueError(f"Invalid fit_threshold={fit_threshold}")

            logging.info(f"Fitting threshold on {fit_threshold} train samples")
            X_train, y_train = self.dataset.sample_n_train_examples(fit_threshold)
            self.llm_clf.fit(X_train, y_train)

        # Evaluate test risk scores
        self._results = evaluate_predictions(
            y_true=y_test.to_numpy(),
            y_pred_scores=self._y_test_scores,
            sensitive_attribute=s_test,
            threshold=self.llm_clf.threshold,
            model_name=self.llm_clf.model_name,
        )

        # Save predictions save path
        self._results["predictions_path"] = test_predictions_save_path.as_posix()

        # Log main results
        msg = (
            f"\n** Test results **\n"
            f"Model: {self.llm_clf.model_name};\n"
            f"\t ECE:       {self._results['ece']:.1%};\n"
            f"\t ROC AUC :  {self.results['roc_auc']:.1%};\n"
            f"\t Accuracy:  {self.results['accuracy']:.1%};\n"
            f"\t Bal. acc.: {self.results['balanced_accuracy']:.1%};\n"
        )
        logging.info(msg)

        # Render plots
        try:
            self.plot_results(show_plots=False)
        except Exception as e:
            logging.error(f"Failed to render plots: {e}")

        # Save results to disk
        self.save_results()

        return self._results

    def plot_results(self, *, show_plots: bool = True):
        """Render evaluation plots and save to disk.

        Parameters
        ----------
        show_plots : bool, optional
            Whether to show plots, by default True.

        Returns
        -------
        plots_paths : dict[str, str]
            The paths to the saved plots.
        """
        if self._results is None:
            raise ValueError("No results to plot. Run the benchmark first.")

        imgs_dir = Path(self.results_dir) / "imgs"
        imgs_dir.mkdir(exist_ok=True, parents=False)
        _, y_test = self.dataset.get_test()

        plots_paths = render_evaluation_plots(
            y_true=y_test.to_numpy(),
            y_pred_scores=self._y_test_scores,
            eval_results=self.results,
            model_name=self.llm_clf.model_name,
            imgs_dir=imgs_dir,
            show_plots=show_plots,
        )

        # Plot fairness plots if sensitive attribute is provided
        if self.task.sensitive_attribute is not None:
            s_test = self.dataset.get_sensitive_attribute_data().loc[y_test.index]

            plots_paths.update(render_fairness_plots(
                y_true=y_test.to_numpy(),
                y_pred_scores=self._y_test_scores,
                sensitive_attribute=s_test,
                eval_results=self.results,
                model_name=self.llm_clf.model_name,
                group_value_map=self.task.sensitive_attribute_value_map(),
                imgs_dir=imgs_dir,
                show_plots=show_plots,
            ))

        self._results["plots"] = plots_paths

        return plots_paths

    def save_results(self, results_root_dir: str | Path = None):
        """Save the benchmark results to disk.

        Parameters
        ----------
        results_root_dir : str | Path, optional
            Path to root directory under which results will be saved. By default
            will use `self.results_root_dir`.
        """
        if self.results is None:
            raise ValueError("No results to save. Run the benchmark first.")

        # Update results directory if provided
        if results_root_dir is not None:
            self.results_root_dir = results_root_dir

        # Save results to disk
        results_file_name = f"results.bench-{hash(self)}.json"
        results_file_path = self.results_dir / results_file_name

        save_json(self.results, path=results_file_path)
        logging.info(f"Saved experiment results to '{results_file_path.as_posix()}'")

    @classmethod
    def make_acs_benchmark(
        cls,
        task_name: str,
        *,
        model: AutoModelForCausalLM | str,
        tokenizer: AutoTokenizer = None,
        data_dir: str | Path = None,
        max_api_rpm: int = None,
        config: BenchmarkConfig = BenchmarkConfig.default_config(),
        **kwargs,
    ) -> Benchmark:
        """Create a standardized calibration benchmark on ACS data.

        Parameters
        ----------
        task_name : str
            The name of the ACS task to use.
        model : AutoModelForCausalLM | str
            The transformers language model to use, or the model ID for a webAPI
            hosted model (e.g., "openai/gpt-4o-mini").
        tokenizer : AutoTokenizer, optional
            The tokenizer used to train the model (if using a transformers
            model). Not required for webAPI models.
        data_dir : str | Path, optional
            Path to the directory to load data from and save data in.
        max_api_rpm : int, optional
            The maximum number of API requests per minute for webAPI models.
        config : BenchmarkConfig, optional
            Extra benchmark configurations, by default will use
            `BenchmarkConfig.default_config()`.
        **kwargs
            Additional arguments passed to `ACSDataset` and `BenchmarkConfig`.
            By default will use a set of standardized configurations for
            reproducibility.

        Returns
        -------
        bench : Benchmark
            The ACS calibration benchmark object.
        """
        # Handle non-standard ACS arguments
        acs_dataset_configs = cls.ACS_DATASET_CONFIGS.copy()
        for arg in acs_dataset_configs:
            if arg in kwargs and kwargs[arg] != cls.ACS_DATASET_CONFIGS[arg]:
                logging.warning(
                    f"Received non-standard ACS argument '{arg}' (using "
                    f"{arg}={kwargs[arg]} instead of default {arg}={cls.ACS_DATASET_CONFIGS[arg]}). "
                    f"This may affect reproducibility.")
                acs_dataset_configs[arg] = kwargs.pop(arg)

        # Update config with any additional kwargs
        config = config.update(**kwargs)

        # Fetch ACS task and dataset
        acs_task = ACSTaskMetadata.get_task(
            name=task_name,
            use_numeric_qa=config.numeric_risk_prompting)

        acs_dataset = ACSDataset.make_from_task(
            task=acs_task,
            cache_dir=data_dir,
            **acs_dataset_configs)

        return cls.make_benchmark(
            task=acs_task,
            dataset=acs_dataset,
            model=model,
            tokenizer=tokenizer,
            max_api_rpm=max_api_rpm,
            config=config,
        )

    @classmethod
    def make_benchmark(
        cls,
        *,
        task: TaskMetadata | str,
        dataset: Dataset,
        model: AutoModelForCausalLM | str,
        tokenizer: AutoTokenizer = None,    # WebAPI models have no local tokenizer
        max_api_rpm: int = None,
        config: BenchmarkConfig = BenchmarkConfig.default_config(),
        **kwargs,
    ) -> Benchmark:
        """Create a calibration benchmark from a given configuration.

        Parameters
        ----------
        task : TaskMetadata | str
            The task metadata object or name of the task to use.
        dataset : Dataset
            The dataset to use for the benchmark.
        model : AutoModelForCausalLM | str
            The transformers language model to use, or the model ID for a webAPI
            hosted model (e.g., "openai/gpt-4o-mini").
        tokenizer : AutoTokenizer, optional
            The tokenizer used to train the model (if using a transformers
            model). Not required for webAPI models.
        max_api_rpm : int, optional
            The maximum number of API requests per minute for webAPI models.
        config : BenchmarkConfig, optional
            Extra benchmark configurations, by default will use
            `BenchmarkConfig.default_config()`.
        **kwargs
            Additional arguments for easier configuration of the benchmark.
            Will simply use these values to update the `config` object.

        Returns
        -------
        bench : Benchmark
            The calibration benchmark object.
        """
        # Update config with any additional kwargs
        config = config.update(**kwargs)

        # Handle TaskMetadata object
        task = TaskMetadata.get_task(task) if isinstance(task, str) else task
        if config.numeric_risk_prompting:
            task.use_numeric_qa = True

        if config.feature_subset is not None and len(config.feature_subset) > 0:
            task = task.create_task_with_feature_subset(config.feature_subset)
            dataset.task = task

        # Check dataset is compatible with task
        if dataset.task is not task and dataset.task.name != task.name:
            raise ValueError(
                f"Dataset task '{dataset.task.name}' does not match the "
                f"provided task '{task.name}'.")

        if config.population_filter is not None:
            dataset = dataset.filter(config.population_filter)

        # Get prompting function
        if config.few_shot:
            print(f"Using few-shot prompting (n={config.few_shot})!")
            encode_row_function = partial(
                encode_row_prompt_few_shot,
                task=task,
                n_shots=config.few_shot,
                dataset=dataset,
                reuse_examples=config.reuse_few_shot_examples,
            )

        else:
            print("Using zero-shot prompting.")
            encode_row_function = partial(encode_row_prompt, task=task)

        # Parse LLMClassifier parameters
        llm_inference_kwargs = {"correct_order_bias": config.correct_order_bias}
        if config.batch_size is not None:
            llm_inference_kwargs["batch_size"] = config.batch_size
        if config.context_size is not None:
            llm_inference_kwargs["context_size"] = config.context_size
        if max_api_rpm is not None and isinstance(model, str):
            llm_inference_kwargs["max_api_rpm"] = max_api_rpm

        # Create LLMClassifier object
        if isinstance(model, str):
            llm_clf = WebAPILLMClassifier(
                model_name=model,
                task=task,
                encode_row=encode_row_function,
                **llm_inference_kwargs,
            )
            logging.info(f"Using webAPI model: {model}")

        else:
            llm_clf = TransformersLLMClassifier(
                model=model,
                tokenizer=tokenizer,
                task=task,
                encode_row=encode_row_function,
                **llm_inference_kwargs,
            )
            logging.info(f"Using local transformers model: {llm_clf.model_name}")

        return cls(
            llm_clf=llm_clf,
            dataset=dataset,
            config=config,
        )



"""Module to hold ACS column mappings from values to natural text.
"""
from functools import partial
from pathlib import Path

from ..col_to_text import ColumnToText
from ..qa_interface import Choice, DirectNumericQA, MultipleChoiceQA
from ._utils import parse_pums_code
from .acs_thresholds import (
    acs_employment_threshold,
    acs_health_insurance_threshold,
    acs_income_threshold,
    acs_mobility_threshold,
    acs_poverty_ratio_threshold,
    acs_public_coverage_threshold,
    acs_travel_time_threshold,
)

# Path to ACS codebook files
ACS_CODEBOOK_DIR = Path(__file__).parent / "data"

# AGEP: Age
acs_age = ColumnToText(
    "AGEP",
    short_description="age",
    value_map=lambda x: f"{int(x)} years old",
)

# COW: Class of Worker
acs_class_of_worker = ColumnToText(
    "COW",
    short_description="class of worker",
    value_map={
        1: "Working for a for-profit private company or organization",
        2: "Working for a non-profit organization",
        3: "Working for the local government",
        4: "Working for the state government",
        5: "Working for the federal government",
        6: "Owner of non-incorporated business, professional practice, or farm",
        7: "Owner of incorporated business, professional practice, or farm",
        8: "Working without pay in a for-profit family business or farm",
        9: "Unemployed and last worked 5 years ago or earlier or never worked",
    },
)

# SCHL: Educational Attainment
acs_schooling = ColumnToText(
    "SCHL",
    short_description="highest educational attainment",
    value_map={
        1: "N/A - no schooling completed",
        2: "Nursery school / preschool",
        3: "Kindergarten",
        4: "1st grade only",
        5: "2nd grade",
        6: "3rd grade",
        7: "4th grade",
        8: "5th grade",
        9: "6th grade",
        10: "7th grade",
        11: "8th grade",
        12: "9th grade",
        13: "10th grade",
        14: "11th grade",
        15: "12th grade, no diploma",
        16: "Regular high school diploma",
        17: "GED or alternative credential",
        18: "Some college, less than 1 year",
        19: "Some college, 1 or more years, no degree",
        20: "Associate's degree",
        21: "Bachelor's degree",
        22: "Master's degree",
        23: "Professional degree beyond a bachelor's degree",
        24: "Doctorate degree",
    },
)

# MAR: Marital Status
acs_marital_status = ColumnToText(
    "MAR",
    short_description="marital status",
    value_map={
        1: "Married",
        2: "Widowed",
        3: "Divorced",
        4: "Separated",
        5: "Never married",
    },
)

# OCCP: Occupation
acs_occupation = ColumnToText(
    "OCCP",
    short_description="occupation",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "OCCP.txt",
        postprocess=lambda x: x[4:].lower().capitalize().strip(),
    ),
)

# POBP: Place of Birth
acs_place_of_birth = ColumnToText(
    "POBP",
    short_description="place of birth",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "POBP.txt",
        postprocess=lambda x: (x[: x.find("/")] if "/" in x else x).strip(),
    ),
)

# RELP: Relationship to Reference Person
acs_relationship = ColumnToText(
    "RELP",
    short_description="relationship to the reference person in the survey",
    value_map={
        0: "The reference person itself",
        1: "Husband/wife",
        2: "Biological son or daughter",
        3: "Adopted son or daughter",
        4: "Stepson or stepdaughter",
        5: "Brother or sister",
        6: "Father or mother",
        7: "Grandchild",
        8: "Parent-in-law",
        9: "Son-in-law or daughter-in-law",
        10: "Other relative",
        11: "Roomer or boarder",
        12: "Housemate or roommate",
        13: "Unmarried partner",
        14: "Foster child",
        15: "Other non-relative",
        16: "Institutionalized group quarters population",
        17: "Non-institutionalized group quarters population",
    },
)

# WKHP: Usual Hours Worked per Week
acs_work_hours = ColumnToText(
    "WKHP",
    short_description="usual number of hours worked per week",
    missing_value_fill="N/A (less than 16 years old, or did not work during the past 12 months)",
    value_map=lambda x: f"{int(x)} hours",
)

# SEX: Sex
acs_sex = ColumnToText(
    "SEX",
    short_description="sex",
    value_map={
        1: "Male",
        2: "Female",
    },
)

# RAC1P: Race
acs_race = ColumnToText(
    "RAC1P",
    short_description="race",
    value_map={
        1: "White",
        2: "Black or African American",
        3: "American Indian",
        4: "Alaska Native",
        5: (
            "American Indian and Alaska Native tribes specified; or American "
            "Indian or Alaska Native, not specified and no other races"),
        6: "Asian",
        7: "Native Hawaiian and Other Pacific Islander",
        8: "Some other race alone (non-White)",
        # 8: "Some other race alone",
        9: "Two or more races",
    },
)

# PINCP: Yearly Income
acs_income = ColumnToText(
    "PINCP",
    short_description="yearly income",
    missing_value_fill="N/A (less than 15 years old)",
    value_map=lambda x: f"${int(x):,}",
)

# PINCP: Yearly Income (Thresholded)
acs_income_qa = MultipleChoiceQA(
    column=acs_income_threshold.apply_to_column_name("PINCP"),
    text="What is this person's estimated yearly income?",
    choices=(
        Choice("Below $50,000", 0),
        Choice("Above $50,000", 1),
    ),
)

acs_income_numeric_qa = DirectNumericQA(
    column=acs_income_threshold.apply_to_column_name("PINCP"),
    text=(
        "What is the probability that this person's estimated yearly income is "
        "above $50,000 ?"
    ),
)

acs_income_target_col = ColumnToText(
    name=acs_income_threshold.apply_to_column_name("PINCP"),
    short_description="yearly income",
    missing_value_fill="N/A (less than 15 years old)",
    question=acs_income_qa,
)

# PUBCOV: Public Health Coverage (Original)
acs_pubcov_og_qa = MultipleChoiceQA(
    column="PUBCOV",
    text="Does this person have public health insurance coverage?",
    choices=(
        Choice("Yes, person is covered by public health insurance", 1),
        Choice("No, person is not covered by public health insurance", 2),  # NOTE: value=2 for no public coverage!
    ),
)

acs_pubcov = ColumnToText(
    "PUBCOV",
    short_description="public health coverage status",
    value_map={
        1: "Covered by public health insurance",
        2: "Not covered by public health insurance",
    },
    question=acs_pubcov_og_qa,
)

# PUBCOV: Public Health Coverage (Thresholded)
acs_pubcov_qa = MultipleChoiceQA(
    column=acs_public_coverage_threshold.apply_to_column_name("PUBCOV"),
    text="Does this person have public health insurance coverage?",
    choices=(
        Choice("Yes, person is covered by public health insurance", 1),
        Choice("No, person is not covered by public health insurance", 0),  # NOTE: value=0 for no public coverage!
    ),
)

acs_pubcov_numeric_qa = DirectNumericQA(
    column=acs_public_coverage_threshold.apply_to_column_name("PUBCOV"),
    text=(
        "What is the probability that this person is covered by public health insurance?"
    ),  # NOTE: value=1 for yes, 0 for no
)

acs_pubcov_target_col = ColumnToText(
    name=acs_public_coverage_threshold.apply_to_column_name("PUBCOV"),
    short_description="public health coverage status",
    question=acs_pubcov_qa,
    use_value_map_only=True,
)

# DIS: Disability Status
acs_disability = ColumnToText(
    "DIS",
    short_description="disability status",
    value_map={
        1: "With a disability",
        2: "No disability",
    },
)

# ESP: Employment Status of Parents
acs_emp_parents = ColumnToText(
    "ESP",
    short_description="employment status of parents",
    value_map={
        1: "Living with two parents, both employed",
        2: "Living with two parents, only Father is employed",
        3: "Living with two parents, only Mother is employed",
        4: "Living with two parents, neither employed",
        5: "Living with Father, and Father is employed",
        6: "Living with Father, and Father is not employed",
        7: "Living with Mother, and Mother is employed",
        8: "Living with Mother, and Mother is not employed",
    },
    missing_value_fill="N/A (not own child of householder, and not child in subfamily)",
)

# CIT: Citizenship Status
acs_citizenship = ColumnToText(
    "CIT",
    short_description="citizenship status",
    value_map={
        1: "Born in the United States",
        2: "Born in Puerto Rico, Guam, the U.S. Virgin Islands, or the Northern Marianas",
        3: "Born abroad of American parents",
        4: "Naturalized US citizen",
        5: "Not a US citizen",
    },
)

# MIG: Mobility Status
acs_mobility = ColumnToText(
    "MIG",
    short_description="mobility status over the last year",
    value_map={
        1: "Lived in the same house one year ago",
        2: "Lived in a different house, outside the United States and Puerto Rico, one year ago",
        3: "Lived in a different house in the United States one year ago",
    },
)

# MIG: Mobility Status (Thresholded)
acs_mobility_qa = MultipleChoiceQA(
    column=acs_mobility_threshold.apply_to_column_name("MIG"),      # NOTE: Thresholded by MIG!=1
    text="Has this person moved in the last year?",
    choices=(
        Choice("No, person has lived in the same house for the last year", 0),
        Choice("Yes, person has moved in the last year", 1),
    ),
)

acs_mobility_numeric_qa = DirectNumericQA(
    column=acs_mobility_threshold.apply_to_column_name("MIG"),      # NOTE: Thresholded by MIG!=1
    text=(
        "What is the probability that this person has moved in the last year?"
    ),  # NOTE: Question should relate to probability of MIG!=1
)

acs_mobility_target_col = ColumnToText(
    name=acs_mobility_threshold.apply_to_column_name("MIG"),
    short_description="mobility status over the last year",
    question=acs_mobility_qa,
    use_value_map_only=True,
)

# MIL: Military Service Status
acs_military = ColumnToText(
    "MIL",
    short_description="military service status",
    value_map={
        1: "Now on active duty",
        2: "On active duty in the past, but not now",
        3: "Only on active duty for training in Reserved/National Guard",
        4: "Never served in the military",
    },
    missing_value_fill="N/A (less than 17 years old)",
)

# ANC: Ancestry
acs_ancestry = ColumnToText(
    "ANC",
    short_description="ancestry",
    value_map={
        1: "Single ancestry",
        2: "Multiple ancestry",
        3: "Unclassified",
        4: "Not reported",
        8: "N/A (information suppressed for certain area codes)",
    },
)

# ANC1P: Detailed Ancestry
acs_detailed_ancestry = ColumnToText(
    "ANC1P",
    short_description="detailed ancestry",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "ANC1P.txt",
        postprocess=lambda x: x.strip(),
    ),
)

# NATIVITY: Nativity
acs_nativity = ColumnToText(
    "NATIVITY",
    short_description="nativity",
    value_map={
        1: "Native",
        2: "Foreign born",
    },
)

# LANX: Language other than English spoken at home
acs_lanx = ColumnToText(
    "LANX",
    short_description="language other than English spoken at home",
    value_map={
        1: "Speaks a language other than English at home.",
        2: "Speaks only English at home.",
    },
    use_value_map_only=True,
    missing_value_fill="N/A (less than 5 years old)",
)

# LANP: Language spoken at home
acs_lanp = ColumnToText(
    "LANP",
    short_description="language spoken at home (other than English)",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "LANP.txt",
        postprocess=lambda x: x.strip(),
    ),
    missing_value_fill="N/A (speaks only English at home)",
)

# ENG: Ability to Speak English
acs_eng = ColumnToText(
    "ENG",
    short_description="ability to speak English",
    value_map={
        1: "Very well",
        2: "Well",
        3: "Not well",
        4: "Not at all",
    },
    missing_value_fill="Native (speaks only English)",
)

# NOP: Nativity of Parents
acs_nop = ColumnToText(
    "NOP",
    short_description="nativity of parents",
    value_map={
        1: "Living with both parents, both of whom are native",
        2: "Living with both parents, Father only is foreign born",
        3: "Living with both parents, Mother only is foreign born",
        4: "Living with both parents, both of whom are foreign born",
        5: "Living with Father only, who is native",
        6: "Living with Father only, who is foreign born",
        7: "Living with Mother only, who is native",
        8: "Living with Mother only, who is foreign born",
    },
    missing_value_fill=(
        "N/A (not own child of householder and not child in subfamily)"
        # OR greater than 17 years old
    ),
)

# DEAR: Hearing Status
acs_hearing = ColumnToText(
    "DEAR",
    short_description="hearing status",
    value_map={
        1: "With hearing difficulty",
        2: "No hearing difficulty",
    },
)

# DEYE: Vision Status
acs_vision = ColumnToText(
    "DEYE",
    short_description="vision status",
    value_map={
        1: "With vision difficulty",
        2: "No vision difficulty",
    },
)

# DREM: Cognitive Status
acs_cognitive = ColumnToText(
    "DREM",
    short_description="cognition status",
    value_map={
        1: "With cognitive difficulty",
        2: "No cognitive difficulty",
    },
    missing_value_fill="N/A (less than 5 years old)",
)

# ESR: Employment Status
acs_employment = ColumnToText(
    "ESR",
    short_description="employment status",
    value_map={
        1: "Civilian employed, at work",
        2: "Civilian employed, with a job but not at work",
        3: "Unemployed",
        4: "Armed forces, at work",
        5: "Armed forces, with a job but not at work",
        6: "Not in labor force",
    },
    missing_value_fill="N/A (less than 16 years old)",
)

# ESR: Employment Status (Thresholded)
acs_employment_qa = MultipleChoiceQA(
    column=acs_employment_threshold.apply_to_column_name("ESR"),
    text="What is this person's employment status?",
    choices=(
        Choice("Employed civilian", 1),
        Choice("Unemployed or in the military", 0),
    ),
)

acs_employment_numeric_qa = DirectNumericQA(
    column=acs_employment_threshold.apply_to_column_name("ESR"),
    text=(
        "What is the probability that this person is an employed civilian?"
        # "What is the probability that this person is currently employed?"
    ),  # NOTE: Question should relate to probability of ESR==1
)

acs_employment_target_col = ColumnToText(
    name=acs_employment_threshold.apply_to_column_name("ESR"),
    short_description="employment status",
    question=acs_employment_qa,
    use_value_map_only=True,
)

# ST: State
acs_state = ColumnToText(
    "ST",
    short_description="resident state",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "ST.txt",
        postprocess=lambda x: x[:x.find("/")].strip(),
    ),
)

# HISP: Hispanic Origin (Detailed)
acs_hispanic = ColumnToText(
    "HISP",
    short_description="Hispanic origin",
    value_map=partial(
        parse_pums_code,
        file=ACS_CODEBOOK_DIR / "HISP.txt",
        postprocess=lambda x: x.strip(),
    ),
)

# FER: Parenthood Status
acs_parenthood = ColumnToText(
    "FER",
    short_description="person has given birth within the last year",
    use_value_map_only=True,
    value_map={
        1: "Person has given birth within the last year.",
        2: "Person has not given birth within the last year.",
    },
    missing_value_fill="N/A (less than 15 years old, or greater than 50 years old, or male)",
)

# JWMNP: Commute Time
acs_commute_time = ColumnToText(
    "JWMNP",
    short_description="commute time",
    value_map=lambda x: f"{int(x)} minutes",
    missing_value_fill="N/A (not a worker, or worker who worked at home)",
)

# JWMNP: Commute Time (Thresholded)
acs_commute_time_qa = MultipleChoiceQA(
    column=acs_travel_time_threshold.apply_to_column_name("JWMNP"),
    text="What is this person's commute time?",
    choices=(
        Choice("Longer than 20 minutes", 1),
        Choice("Less than 20 minutes", 0),
    ),
)

acs_commute_time_numeric_qa = DirectNumericQA(
    column=acs_travel_time_threshold.apply_to_column_name("JWMNP"),
    text=(
        "What is the probability that this person's commute time is longer than 20 minutes?"
    ),  # NOTE: Question should relate to probability of JWMNP>20
)

acs_travel_time_target_col = ColumnToText(
    name=acs_travel_time_threshold.apply_to_column_name("JWMNP"),
    short_description="commute time",
    question=acs_commute_time_qa,
    use_value_map_only=True,
)

# JWTR: Commute Method
acs_commute_method = ColumnToText(
    "JWTR",
    short_description="means of transportation to work",
    value_map={
        1: "Car, truck, or van",
        2: "Bus or trolley bus",
        3: "Streetcar or trolley car",
        4: "Subway or elevated",
        5: "Railroad",
        6: "Ferryboat",
        7: "Taxicab",
        8: "Motorcycle",
        9: "Bicycle",
        10: "Walked",
        11: "Worked at home",
        12: "Other method",
    },
)


def describe_income_to_poverty_ratio_pct(ratio_pct: float) -> str:
    """Describe the income-to-poverty ratio in natural language."""
    context_msg: str

    if ratio_pct < 90:
        context_msg = "below the poverty line"
    elif 90 <= ratio_pct < 110:
        context_msg = "at or near the poverty line"
    else:
        context_msg = "above the poverty line"

    return f"{int(ratio_pct)}% of the poverty line income, which is {context_msg}"


# POVPIP: Income-to-Poverty Ratio
acs_poverty_ratio = ColumnToText(
    "POVPIP",
    short_description="income-to-poverty ratio",
    value_map=describe_income_to_poverty_ratio_pct,
)

# POVPIP: Income-to-Poverty Ratio (Thresholded)
acs_poverty_ratio_qa = MultipleChoiceQA(
    column=acs_poverty_ratio_threshold.apply_to_column_name("POVPIP"),
    text=(
        "Is this person's income-to-poverty ratio below 2.5 ? "
        "That is, is this person's annual income below 2.5 times the poverty line income?",
    ),
    choices=(
        Choice("Yes, this person earns below 2.5 times the poverty line income", 1),
        Choice("No, this person earns above 2.5 times the poverty line income.", 0),
    ),
)

acs_poverty_ratio_target_col = ColumnToText(
    name=acs_poverty_ratio_threshold.apply_to_column_name("POVPIP"),
    short_description="income-to-poverty ratio is below 2.5",
    question=acs_poverty_ratio_qa,
    use_value_map_only=True,
)

# GCL: Grandparent Living with Grandchildren
acs_gcl_col = ColumnToText(
    "GCL",
    short_description="grandparent living with grandchildren",
    use_value_map_only=True,
    value_map={
        1: "Household includes grandparent living with grandchildren",
        2: "Household does not include grandparents living with grandchildren",
    },
    missing_value_fill="N/A (less than 30 years old, or living in institutional group quarters)",
)

# PUMA: Public Use Microdata Area Code
# TODO: assign meaningful natural-text values to PUMA codes
# > Consult: https://www2.census.gov/geo/pdfs/reference/puma/2010_PUMA_Names.pdf
acs_puma_col = ColumnToText(
    "PUMA",
    short_description="Public Use Microdata Area (PUMA) code",
    use_value_map_only=True,
    value_map=lambda x: f"Public Use Microdata Area (PUMA) code: {int(x)}.",
    # missing_value_fill="N/A (less than 16 years old)",
)

# POWPUMA: Place of Work PUMA
acs_powpuma_col = ColumnToText(
    "POWPUMA",
    short_description="place of work PUMA",
    use_value_map_only=True,
    value_map=lambda x: f"Public Use Microdata Area (PUMA) code for the place of work: {int(x)}.",
    # missing_value_fill="N/A (not a worker, or worker who worked at home)",
)

# HINS2: Health Insurance Coverage through Private Company
acs_health_ins_2_col = ColumnToText(
    "HINS2",
    short_description="acquired health insurance directly from an insurance company",
    use_value_map_only=True,
    value_map={
        1: "Person has purchased insurance directly from an insurance company",
        2: (
            "Person has not purchased insurance directly from an insurance "
            "company (is either uninsured or insured through another source)",
        )
    },
)

# HINS2: Health Insurance Coverage through Private Company (Thresholded)
acs_health_ins_2_qa = MultipleChoiceQA(
    column=acs_health_insurance_threshold.apply_to_column_name("HINS2"),
    text="Has this person purchased health insurance directly from an insurance company?",
    choices=(
        Choice("Yes, this person has health insurance through a private company", 1),
        Choice("No, this person either has insurance through other means or is uninsured", 0),
    ),
)

acs_health_ins_2_numeric_qa = DirectNumericQA(
    column=acs_health_insurance_threshold.apply_to_column_name("HINS2"),
    text=(
        "What is the probability that this person has purchased health "
        "insurance directly through a private company?"
    ),  # NOTE: Question should relate to probability of HINS2==1
)

acs_health_ins_2_target_col = ColumnToText(
    name=acs_health_insurance_threshold.apply_to_column_name("HINS2"),
    short_description="acquired health insurance directly from an insurance company",
    question=acs_health_ins_2_qa,
    use_value_map_only=True,
)
